//! Abstract layer for Semtech SX1276/77/78/79.

use core::fmt;

/// The error returned from [`Conn::send`].
///
/// [`Conn::Send`]: trait.Conn.html#method.send
pub struct SendError<C: Conn> {
  /// Connection.
  pub tokens: C,
  /// SPI interface.
  pub spi: C::Spi,
  /// Kind of the error.
  pub kind: C::SendErrorKind,
}

/// The error returned from [`Reset::reset`].
///
/// [`Reset::reset`]: trait.Reset.html#method.reset
pub struct ResetError<C: Reset<T>, T> {
  /// Connection.
  pub tokens: C,
  /// Timer.
  pub timer: T,
  /// Kind of the error.
  pub kind: C::ResetErrorKind,
}

/// Generic Semtech SX1276/77/78/79.
pub trait Conn: Sized + Send + 'static {
  /// The SPI interface.
  type Spi: Sized + Send + 'static;

  /// Erroneous result of the future returned by [`send`].
  ///
  /// [`send`]: #method.send
  type SendErrorKind: Sized + Send + PartialEq + Eq + fmt::Debug + 'static;

  /// Initializes inner peripherals.
  fn init(&self);

  /// Makes SPI transaction.
  unsafe fn transfer(
    self,
    spi: Self::Spi,
    address: u8,
    count: usize,
    tx_buf: *const u8,
    rx_buf: *mut u8,
  ) -> Box<Future<Item = (Self, Self::Spi), Error = SendError<Self>> + Send>;
}

/// Generic reset over timer `T`.
pub trait Reset<T>: Conn {
  /// Erroneous result of the future returned by [`reset`].
  ///
  /// [`reset`]: #method.reset
  type ResetErrorKind: Sized + Send + PartialEq + Eq + fmt::Debug + 'static;

  /// Runs a reset procedure, and returns a future for its result.
  fn reset(
    self,
    timer: T,
  ) -> Box<Future<Item = (Self, T), Error = ResetError<Self, T>> + Send>;
}

impl<C: Conn> fmt::Debug for SendError<C> {
  #[inline(always)]
  fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
    self.kind.fmt(f)
  }
}

impl<C: Reset<T>, T> fmt::Debug for ResetError<C, T> {
  #[inline(always)]
  fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
    self.kind.fmt(f)
  }
}

impl<C: Conn> PartialEq for SendError<C> {
  #[inline(always)]
  fn eq(&self, other: &Self) -> bool {
    self.kind.eq(&other.kind)
  }
}

impl<C: Reset<T>, T> PartialEq for ResetError<C, T> {
  #[inline(always)]
  fn eq(&self, other: &Self) -> bool {
    self.kind.eq(&other.kind)
  }
}

impl<C: Conn> Eq for SendError<C> {}

impl<C: Reset<T>, T> Eq for ResetError<C, T> {}
