//! FSK/OOK mode registers.

use drone_core::reg::RegVal;
use drone_sx1276_macros::spi_reg;
use mode::FskOok;
use reg::{RSpiReg, SpiReg, WSpiReg};
use reg::never::NeverRegFskOok;

spi_reg! {
  /// Operating mode & LoRa™ / FSK selection.
  OpMode;
  BitrateMsb;
  FskOok;
  0x01 0x01 RSpiReg WSpiReg;
  /// LoRa™ / FSK selection.
  long_range_mode { 7 1 }
  /// Modulation scheme.
  modulation_type { 5 2 }
  /// Access Low Frequency Mode registers.
  low_frequency_mode_on { 3 1 }
  /// Transceiver mode.
  mode { 0 3 }
}

spi_reg! {
  /// Bit Rate setting, Most Significant Bits.
  BitrateMsb;
  BitrateLsb;
  FskOok;
  0x02 0x1A RSpiReg WSpiReg;
  /// MSB of Bit Rate.
  bit_rate_15_8 { 0 8 }
}

spi_reg! {
  /// Bit Rate setting, Least Significant Bits.
  BitrateLsb;
  FdevMsb;
  FskOok;
  0x03 0x0B RSpiReg WSpiReg;
  /// LSB of bit rate.
  bit_rate_7_0 { 0 8 }
}

spi_reg! {
  /// Frequency Deviation setting, Most Significant Bits.
  FdevMsb;
  FdevLsb;
  FskOok;
  0x04 0x00 RSpiReg WSpiReg;
  /// MSB of the frequency deviation.
  fdev_13_8 { 0 6 }
}

spi_reg! {
  /// Frequency Deviation setting, Least Significant Bits.
  FdevLsb;
  FrfMsb;
  FskOok;
  0x05 0x52 RSpiReg WSpiReg;
  /// LSB of the frequency deviation.
  fdev_7_0 { 0 8 }
}

spi_reg! {
  /// RF Carrier Frequency, Most Significant Bits.
  FrfMsb;
  FrfMid;
  FskOok;
  0x06 0x6C RSpiReg WSpiReg;
  /// MSB of the RF carrier frequency.
  frf_23_16 { 0 8 }
}

spi_reg! {
  /// RF Carrier Frequency, Intermediate Bits.
  FrfMid;
  FrfLsb;
  FskOok;
  0x07 0x80 RSpiReg WSpiReg;
  /// MSB of the RF carrier frequency.
  frf_15_8 { 0 8 }
}

spi_reg! {
  /// RF Carrier Frequency, Least Significant Bits.
  FrfLsb;
  PaConfig;
  FskOok;
  0x08 0x00 RSpiReg WSpiReg;
  /// LSB of RF carrier frequency.
  frf_7_0 { 0 8 }
}

spi_reg! {
  /// PA selection and Output Power control.
  PaConfig;
  PaRamp;
  FskOok;
  0x09 0x4F RSpiReg WSpiReg;
  /// Selects PA output pin.
  pa_select { 7 1 }
  /// Select max output power.
  max_power { 4 3 }
  /// Select output power.
  output_power { 0 4 }
}

spi_reg! {
  /// Control of PA ramp time, low phase noise PLL.
  PaRamp;
  Ocp;
  FskOok;
  0x0A 0x09 RSpiReg WSpiReg;
  /// Data shaping.
  modulation_shaping { 5 2 }
  /// Rise/Fall time of ramp up/down in FSK.
  pa_ramp { 0 4 }
}

spi_reg! {
  /// Over Current Protection control.
  Ocp;
  Lna;
  FskOok;
  0x0B 0x2B RSpiReg WSpiReg;
  /// Enables overload current protection (OCP) for the PA.
  ocp_on { 5 1 }
  /// Trimming of OCP current.
  ocp_trim { 0 5 }
}

spi_reg! {
  /// LNA settings.
  Lna;
  RxConfig;
  FskOok;
  0x0C 0x20 RSpiReg WSpiReg;
  /// LNA gain setting.
  lna_gain { 5 3 }
  /// Low Frequency (RFI_LF) LNA current adjustment.
  lna_boost_lf { 3 2 }
  /// High Frequency (RFI_HF) LNA current adjustment.
  lna_boost_hf { 0 2 }
}

spi_reg! {
  /// AFC, AGC, ctrl.
  RxConfig;
  RssiConfig;
  FskOok;
  0x0D 0x08 RSpiReg WSpiReg;
  /// Turns on the mechanism restarting the receiver automatically if it gets
  /// saturated or a packet collision is detected.
  restart_rx_on_collision { 7 1 }
  /// Triggers a manual Restart of the Receiver chain when set to 1.
  restart_rx_without_pll_lock { 6 1 }
  /// Triggers a manual Restart of the Receiver chain when set to 1.
  restart_rx_with_pll_lock { 5 1 }
  /// AFC.
  afc_auto_on { 4 1 }
  /// AGC.
  agc_auto_on { 3 1 }
  /// Selects the event triggering AGC and/or AFC at receiver startup.
  rx_trigger { 0 3 }
}

spi_reg! {
  /// RSSI.
  RssiConfig;
  RssiCollision;
  FskOok;
  0x0E 0x02 RSpiReg WSpiReg;
  /// Signed RSSI offset, to compensate for the possible losses/gains in the
  /// front-end (LNA, SAW filter...).
  rssi_offset { 3 5 }
  /// Defines the number of samples taken to average the RSSI result.
  rssi_smoothing { 0 3 }
}

spi_reg! {
  /// RSSI Collision detector.
  RssiCollision;
  RssiThresh;
  FskOok;
  0x0F 0x0A RSpiReg WSpiReg;
  /// Sets the threshold used to consider that an interferer is detected,
  /// witnessing a packet collision.
  rssi_collision_threshold { 0 8 }
}

spi_reg! {
  /// RSSI Threshold control.
  RssiThresh;
  RssiValue;
  FskOok;
  0x10 0xFF RSpiReg WSpiReg;
  /// RSSI trigger level for the Rssi interrupt.
  rssi_threshold { 0 8 }
}

spi_reg! {
  /// RSSI value in dBm.
  RssiValue;
  RxBw;
  FskOok;
  0x11 0 RSpiReg;
  /// Absolute value of the RSSI in dBm.
  rssi_value { 0 8 }
}

spi_reg! {
  /// Channel Filter BW Control.
  RxBw;
  AfcBw;
  FskOok;
  0x12 0x15 RSpiReg WSpiReg;
  /// Channel filter bandwidth control.
  rx_bw_mant { 3 2 }
  /// Channel filter bandwidth control.
  rx_bw_exp { 0 3 }
}

spi_reg! {
  /// AFC Channel Filter BW.
  AfcBw;
  OokPeak;
  FskOok;
  0x13 0x0B RSpiReg WSpiReg;
  /// RxBwMant parameter used during the AFC.
  rx_bw_mant_afc { 3 2 }
  /// RxBwExp parameter used during the AFC.
  rx_bw_exp_afc { 0 3 }
}

spi_reg! {
  /// OOK demodulator.
  OokPeak;
  OokFix;
  FskOok;
  0x14 0x28 RSpiReg WSpiReg;
  /// Enables the Bit Synchronizer.
  bit_sync_on { 5 1 }
  /// Selects the type of threshold in the OOK data slicer.
  ook_thresh_type { 3 2 }
  /// Size of each decrement of the RSSI threshold in the OOK demodulator.
  ook_peak_thesh_step { 0 3 }
}

spi_reg! {
  /// Threshold of the OOK demod.
  OokFix;
  OokAvg;
  FskOok;
  0x15 0x0C RSpiReg WSpiReg;
  /// Fixed threshold for the Data Slicer in OOK mode.
  ook_fixed_threshold { 0 8 }
}

spi_reg! {
  /// Average of the OOK demod.
  OokAvg;
  NeverRegFskOok;
  FskOok;
  0x16 0x12 RSpiReg WSpiReg;
  /// Period of decrement of the RSSI threshold in the OOK demodulator.
  ook_peak_thresh_dec { 5 3 }
  /// Static offset added to the threshold in average mode in order to reduce
  /// glitching activity (OOK only).
  ook_average_offset { 2 2 }
  /// Filter coefficients in average mode of the OOK demodulator.
  ook_average_thresh_filt { 0 2 }
}

spi_reg! {
  /// AFC and FEI control.
  AfcFei;
  AfcMsb;
  FskOok;
  0x1A 0x00 RSpiReg WSpiReg;
  /// Triggers an AGC sequence when set to 1.
  agc_start { 4 1 }
  /// Clear AFC register set in Rx mode. Always reads 0.
  afc_clear { 1 1 }
  /// Only valid if AfcAutoOn is set.
  afc_auto_clear_on { 0 1 }
}

spi_reg! {
  /// Frequency correction value of the AFC.
  AfcMsb;
  AfcLsb;
  FskOok;
  0x1B 0x00 RSpiReg WSpiReg;
  /// MSB of the AfcValue, 2's complement format.
  afc_value_15_8 { 0 8 }
}

spi_reg! {
  /// Frequency correction value of the AFC.
  AfcLsb;
  FeiMsb;
  FskOok;
  0x1C 0x00 RSpiReg WSpiReg;
  /// LSB of the AfcValue, 2's complement format.
  afc_value_7_0 { 0 8 }
}

spi_reg! {
  /// Value of the calculated frequency error.
  FeiMsb;
  FeiLsb;
  FskOok;
  0x1D 0x00 RSpiReg WSpiReg;
  /// MSB of the measured frequency offset, 2's complement.
  fei_value_15_8 { 0 8 }
}

spi_reg! {
  /// Value of the calculated frequency error.
  FeiLsb;
  PreambleDetect;
  FskOok;
  0x1E 0x00 RSpiReg WSpiReg;
  /// LSB of the measured frequency offset, 2's complement.
  fei_value_7_0 { 0 8 }
}

spi_reg! {
  /// Settings of the Preamble Detector.
  PreambleDetect;
  RxTimeout1;
  FskOok;
  0x1F 0x40 RSpiReg WSpiReg;
  /// Enables Preamble detector when set to 1.
  preamble_detector_on { 7 1 }
  /// Number of Preamble bytes to detect to trigger an interrupt.
  preamble_detector_size { 5 2 }
  /// Number or chip errors tolerated over PreambleDetectorSize.
  preamble_detector_tol { 0 5 }
}

spi_reg! {
  /// Timeout Rx request and RSSI.
  RxTimeout1;
  RxTimeout2;
  FskOok;
  0x20 0x00 RSpiReg WSpiReg;
  /// Timeout interrupt is generated TimeoutRxRssi*16*Tbit after switching to Rx
  /// mode if Rssi interrupt doesn't occur.
  timeout_rx_rssi { 0 8 }
}

spi_reg! {
  /// Timeout RSSI and PayloadReady.
  RxTimeout2;
  RxTimeout3;
  FskOok;
  0x21 0x00 RSpiReg WSpiReg;
  /// Timeout interrupt is generated TimeoutRxPreamble*16*Tbit after switching
  /// to Rx mode if Preamble interrupt doesn't occur.
  timeout_rx_preamble { 0 8 }
}

spi_reg! {
  /// Timeout RSSI and SyncAddress.
  RxTimeout3;
  RxDelay;
  FskOok;
  0x22 0x00 RSpiReg WSpiReg;
  /// Timeout interrupt is generated TimeoutSignalSync*16*Tbit after the Rx mode
  /// is programmed, if SyncAddress doesn't occur.
  timeout_signal_sync { 0 8 }
}

spi_reg! {
  /// Delay between Rx cycles.
  RxDelay;
  Osc;
  FskOok;
  0x23 0x00 RSpiReg WSpiReg;
  /// Additional delay before an automatic receiver restart is launched.
  inter_packet_rx_delay { 0 8 }
}

spi_reg! {
  /// RC Oscillators Settings, CLKOUT frequency.
  Osc;
  PreambleMsb;
  FskOok;
  0x24 0x05 RSpiReg WSpiReg;
  /// Triggers the calibration of the RC oscillator when set.
  rc_cal_start { 3 1 }
  /// Selects CLKOUT frequency.
  clk_out { 0 3 }
}

spi_reg! {
  /// Preamble length, MSB.
  PreambleMsb;
  PreambleLsb;
  FskOok;
  0x25 0x00 RSpiReg WSpiReg;
  /// Size of the preamble to be sent.
  preamble_size_15_8 { 0 8 }
}

spi_reg! {
  /// Preamble length, LSB.
  PreambleLsb;
  SyncConfig;
  FskOok;
  0x26 0x03 RSpiReg WSpiReg;
  /// Size of the preamble to be sent.
  preamble_size_7_0 { 0 8 }
}

spi_reg! {
  /// Sync Word Recognition control.
  SyncConfig;
  SyncValue1;
  FskOok;
  0x27 0x93 RSpiReg WSpiReg;
  /// Controls the automatic restart of the receiver after the reception of a
  /// valid packet (PayloadReady or CrcOk).
  auto_restart_rx_mode { 6 2 }
  /// Sets the polarity of the Preamble.
  preamble_polarity { 5 1 }
  /// Enables the Sync word generation and detection.
  sync_on { 4 1 }
  /// Size of the Sync word.
  sync_size { 0 3 }
}

spi_reg! {
  /// Sync Word bytes 1.
  SyncValue1;
  SyncValue2;
  FskOok;
  0x28 0x55 RSpiReg WSpiReg;
  /// 1st byte of Sync word. (MSB byte)
  sync_value_63_56 { 0 8 }
}

spi_reg! {
  /// Sync Word bytes 2.
  SyncValue2;
  SyncValue3;
  FskOok;
  0x29 0x55 RSpiReg WSpiReg;
  /// 2nd byte of Sync word.
  sync_value_55_48 { 0 8 }
}

spi_reg! {
  /// Sync Word bytes 3.
  SyncValue3;
  SyncValue4;
  FskOok;
  0x2A 0x55 RSpiReg WSpiReg;
  /// 3rd byte of Sync word.
  sync_value_47_40 { 0 8 }
}

spi_reg! {
  /// Sync Word bytes 4.
  SyncValue4;
  SyncValue5;
  FskOok;
  0x2B 0x55 RSpiReg WSpiReg;
  /// 4th byte of Sync word.
  sync_value_39_32 { 0 8 }
}

spi_reg! {
  /// Sync Word bytes 5.
  SyncValue5;
  SyncValue6;
  FskOok;
  0x2C 0x55 RSpiReg WSpiReg;
  /// 5th byte of Sync word.
  sync_value_31_24 { 0 8 }
}

spi_reg! {
  /// Sync Word bytes 6.
  SyncValue6;
  SyncValue7;
  FskOok;
  0x2D 0x55 RSpiReg WSpiReg;
  /// 6th byte of Sync word.
  sync_value_23_16 { 0 8 }
}

spi_reg! {
  /// Sync Word bytes 7.
  SyncValue7;
  SyncValue8;
  FskOok;
  0x2E 0x55 RSpiReg WSpiReg;
  /// 7th byte of Sync word.
  sync_value_15_8 { 0 8 }
}

spi_reg! {
  /// Sync Word bytes 8.
  SyncValue8;
  PacketConfig1;
  FskOok;
  0x2F 0x55 RSpiReg WSpiReg;
  /// 8th byte of Sync word.
  sync_value_7_0 { 0 8 }
}

spi_reg! {
  /// Packet mode settings.
  PacketConfig1;
  PacketConfig2;
  FskOok;
  0x30 0x90 RSpiReg WSpiReg;
  /// Defines the packet format used.
  packet_format { 7 1 }
  /// Defines DC-free encoding/decoding performed.
  dc_free { 5 2 }
  /// Enables CRC calculation/check (Tx/Rx).
  crc_on { 4 1 }
  /// Defines the behavior of the packet handler when CRC check fails.
  crc_auto_clear_off { 3 1 }
  /// Defines address based filtering in Rx.
  address_filtering { 1 2 }
  /// Selects the CRC and whitening algorithms.
  crc_whitening_type { 0 1 }
}

spi_reg! {
  /// Packet mode settings.
  PacketConfig2;
  PayloadLength;
  FskOok;
  0x31 0x40 RSpiReg WSpiReg;
  /// Data processing mode.
  data_mode { 6 1 }
  /// Enables the io-homecontrol® compatibility mode.
  io_home_on { 5 1 }
  /// Enables the Beacon mode in Fixed packet format.
  beacon_on { 3 1 }
  /// Packet Length Most significant bits.
  payload_length_10_8 { 0 3 }
}

spi_reg! {
  /// Payload length setting.
  PayloadLength;
  NodeAdrs;
  FskOok;
  0x32 0x40 RSpiReg WSpiReg;
  /// Payload length.
  payload_length_7_0 { 0 8 }
}

spi_reg! {
  /// Node address.
  NodeAdrs;
  BroadcastAdrs;
  FskOok;
  0x33 0x00 RSpiReg WSpiReg;
  /// Node address used in address filtering.
  node_address { 0 8 }
}

spi_reg! {
  /// Broadcast address.
  BroadcastAdrs;
  FifoThresh;
  FskOok;
  0x34 0x00 RSpiReg WSpiReg;
  /// Broadcast address used in address filtering.
  broadcast_address { 0 8 }
}

spi_reg! {
  /// Fifo threshold, Tx start condition.
  FifoThresh;
  SeqConfig1;
  FskOok;
  0x35 0x0F RSpiReg WSpiReg;
  /// Defines the condition to start packet transmission.
  tx_start_condition { 7 1 }
  /// Used to trigger FifoLevel interrupt.
  fifo_threshold { 0 6 }
}

spi_reg! {
  /// Top level Sequencer settings.
  SeqConfig1;
  SeqConfig2;
  FskOok;
  0x36 0x00 RSpiReg WSpiReg;
  /// Controls the top level Sequencer.
  sequencer_start { 7 1 }
  /// Forces the Sequencer Off.
  sequencer_stop { 6 1 }
  /// Selects chip mode during the state.
  idle_mode { 5 1 }
  /// Controls the Sequencer transition when SequencerStart is set to 1 in Sleep
  /// or Standby mode.
  from_start { 3 2 }
  /// Selects the Sequencer LowPower state after a to LowPowerSelection
  /// transition.
  low_power_selection { 2 1 }
  /// Controls the Sequencer transition from the Idle state on a T1 interrupt.
  from_idle { 1 1 }
  /// Controls the Sequencer transition from the Transmit state.
  from_transmit { 0 1 }
}

spi_reg! {
  /// Top level Sequencer settings.
  SeqConfig2;
  TimerResol;
  FskOok;
  0x37 0x00 RSpiReg WSpiReg;
  /// Controls the Sequencer transition from the Receive state.
  from_receive { 5 3 }
  /// Controls the state-machine transition from the Receive state on a
  /// RxTimeout interrupt (and on PayloadReady if FromReceive = 011).
  from_rx_timeout { 3 2 }
  /// Controls the state-machine transition from the PacketReceived state.
  from_packet_received { 0 3 }
}

spi_reg! {
  /// Timer 1 and 2 resolution control.
  TimerResol;
  Timer1Coef;
  FskOok;
  0x38 0x00 RSpiReg WSpiReg;
  /// Resolution of Timer 1.
  timer_1_resolution { 2 2 }
  /// Resolution of Timer 2.
  timer_2_resolution { 0 2 }
}

spi_reg! {
  /// Timer 1 setting.
  Timer1Coef;
  Timer2Coef;
  FskOok;
  0x39 0xF5 RSpiReg WSpiReg;
  /// Multiplying coefficient for Timer 1.
  timer_1_coefficient { 0 8 }
}

spi_reg! {
  /// Timer 2 setting.
  Timer2Coef;
  ImageCal;
  FskOok;
  0x3A 0x20 RSpiReg WSpiReg;
  /// Multiplying coefficient for Timer 2.
  timer_2_coefficient { 0 8 }
}

spi_reg! {
  /// Image calibration engine control.
  ImageCal;
  Temp;
  FskOok;
  0x3B 0x82 RSpiReg WSpiReg;
  /// Controls the Image calibration mechanism.
  auto_image_cal_on { 7 1 }
  /// Triggers the IQ and RSSI calibration when set in Standby mode.
  image_cal_start { 6 1 }
  /// Set to 1 while the Image and RSSI calibration are running. Toggles back to
  /// 0 when the process is completed.
  image_cal_running { 5 1 }
  /// IRQ flag witnessing a temperature change exceeding TempThreshold since the
  /// last Image and RSSI calibration.
  temp_change { 3 1 }
  /// Temperature change threshold to trigger a new I/Q calibration.
  temp_threshold { 1 2 }
  /// Controls the temperature monitor operation.
  temp_monitor_off { 0 1 }
}

spi_reg! {
  /// Temperature Sensor value.
  Temp;
  LowBat;
  FskOok;
  0x3C 0 RSpiReg;
  /// Measured temperature.
  temp_value { 0 8 }
}

spi_reg! {
  /// Low Battery Indicator Settings.
  LowBat;
  IrqFlags1;
  FskOok;
  0x3D 0x02 RSpiReg WSpiReg;
  /// Low Battery detector enable signal.
  low_bat_on { 3 1 }
  /// Trimming of the LowBat threshold.
  low_bat_trim { 0 3 }
}

spi_reg! {
  /// Status register: PLL Lock state, Timeout, RSSI.
  IrqFlags1;
  IrqFlags2;
  FskOok;
  0x3E 0x80 RSpiReg WSpiReg;
  /// Set when the operation mode is ready. Cleared when changing the operating
  /// mode.
  mode_ready { 7 1 }
  /// Set in Rx mode, after RSSI, AGC and AFC. Cleared when leaving Rx.
  rx_ready { 6 1 }
  /// Set in Tx mode, after PA ramp-up. Cleared when leaving Tx.
  tx_ready { 5 1 }
  /// Set (in FS, Rx or Tx) when the PLL is locked. Cleared when it is not.
  pll_lock { 4 1 }
  /// Set in Rx when the RssiValue exceeds RssiThreshold. Cleared when leaving
  /// Rx or setting this bit to 1.
  rssi { 3 1 }
  /// Set when a timeout occurs. Cleared when leaving Rx or FIFO is emptied.
  timeout { 2 1 }
  /// Set when the Preamble Detector has found valid Preamble. Bit clear when
  /// set to 1.
  preamble_detect { 1 1 }
  /// Set when Sync and Address (if enabled) are detected. Cleared when leaving
  /// Rx or FIFO is emptied.
  sync_address_match { 0 1 }
}

spi_reg! {
  /// Status register: FIFO handling flags, Low Battery.
  IrqFlags2;
  DioMapping1;
  FskOok;
  0x3F 0x40 RSpiReg WSpiReg;
  /// Set when FIFO is full (i.e. contains 66 bytes), else cleared.
  fifo_full { 7 1 }
  /// Set when FIFO is empty, and cleared when there is at least 1 byte in the
  /// FIFO.
  fifo_empty { 6 1 }
  /// Set when the number of bytes in the FIFO strictly exceeds FifoThreshold,
  /// else cleared.
  fifo_level { 5 1 }
  /// Set when FIFO overrun occurs. (except in Sleep mode) Flag(s) and FIFO are
  /// cleared when this bit is set. The FIFO then becomes immediately available
  /// for the next transmission/reception.
  fifo_overrun { 4 1 }
  /// Set in Tx when the complete packet has been sent. Cleared when exiting Tx.
  packet_sent { 3 1 }
  /// Set in Rx when the payload is ready (i.e. last byte received and CRC, if
  /// enabled and CrcAutoClearOff is cleared, is Ok). Cleared when FIFO is
  /// empty.
  payload_ready { 2 1 }
  /// Set in Rx when the CRC of the payload is Ok. Cleared when FIFO is empty.
  crc_ok { 1 1 }
  /// Set when the battery voltage drops below the Low Battery threshold.
  /// Cleared only when set to 1 by the user.
  low_bat { 0 1 }
}

spi_reg! {
  /// Mapping of pins DIO0 to DIO3.
  DioMapping1;
  DioMapping2;
  FskOok;
  0x40 0x00 RSpiReg WSpiReg;
  /// Mapping of pin DIO0.
  dio0_mapping { 6 2 }
  /// Mapping of pin DIO1.
  dio1_mapping { 4 2 }
  /// Mapping of pin DIO2.
  dio2_mapping { 2 2 }
  /// Mapping of pin DIO3.
  dio3_mapping { 0 2 }
}

spi_reg! {
  /// Mapping of pins DIO4 and DIO5, ClkOut frequency.
  DioMapping2;
  Version;
  FskOok;
  0x41 0x00 RSpiReg WSpiReg;
  /// Mapping of pin DIO4.
  dio4_mapping { 6 2 }
  /// Mapping of pin DIO5.
  dio5_mapping { 4 2 }
  /// Allows the mapping of either Rssi Or PreambleDetect to the DIO pins.
  preamble_detect { 0 1 }
}

spi_reg! {
  /// Semtech ID relating the silicon revision.
  Version;
  NeverRegFskOok;
  FskOok;
  0x42 0x12 RSpiReg;
  /// Version code of the chip. Bits 7-4 give the full revision number; bits 3-0
  /// give the metal mask revision number.
  version { 0 8 }
}

spi_reg! {
  /// Control the fast frequency hopping mode.
  PllHop;
  NeverRegFskOok;
  FskOok;
  0x44 0x2D RSpiReg WSpiReg;
  /// Bypasses the main state machine for a quick frequency hop.
  fast_hop_on { 7 1 }
}

spi_reg! {
  /// TCXO or XTAL input setting.
  Tcxo;
  NeverRegFskOok;
  FskOok;
  0x4B 0x09 RSpiReg WSpiReg;
  /// Controls the crystal oscillator.
  tcxo_input_on { 4 1 }
}

spi_reg! {
  /// Higher power settings of the PA.
  PaDac;
  NeverRegFskOok;
  FskOok;
  0x4D 0x84 RSpiReg WSpiReg;
  /// Enables the +20dBm option on PA_BOOST pin.
  pa_dac { 0 3 }
}

spi_reg! {
  /// Stored temperature during the former IQ Calibration.
  FormerTemp;
  NeverRegFskOok;
  FskOok;
  0x5B 0 RSpiReg WSpiReg;
  /// Temperature saved during the latest IQ (RSSI and Image) calibration.
  former_temp { 0 8 }
}

spi_reg! {
  /// Fractional part in the Bit Rate division ratio.
  BitRateFrac;
  NeverRegFskOok;
  FskOok;
  0x5D 0x00 RSpiReg WSpiReg;
  /// Fractional part of the bit rate divider (Only valid for FSK).
  bit_rate_frac { 0 4 }
}

spi_reg! {
  /// Adjustment of the AGC thresholds.
  AgcRef;
  AgcThresh1;
  FskOok;
  0x61 0x13 RSpiReg WSpiReg;
  /// Sets the floor reference for all AGC thresholds.
  agc_reference_level { 0 6 }
}

spi_reg! {
  /// Adjustment of the AGC thresholds.
  AgcThresh1;
  AgcThresh2;
  FskOok;
  0x62 0x0E RSpiReg WSpiReg;
  /// Defines the 1st AGC Threshold.
  agc_step_1 { 0 5 }
}

spi_reg! {
  /// Adjustment of the AGC thresholds.
  AgcThresh2;
  AgcThresh3;
  FskOok;
  0x63 0x5B RSpiReg WSpiReg;
  /// Defines the 2nd AGC Threshold.
  agc_step_2 { 4 4 }
  /// Defines the 3rd AGC Threshold.
  agc_step_3 { 0 4 }
}

spi_reg! {
  /// Adjustment of the AGC thresholds.
  AgcThresh3;
  NeverRegFskOok;
  FskOok;
  0x64 0xDB RSpiReg WSpiReg;
  /// Defines the 4th AGC Threshold.
  agc_step_4 { 4 4 }
  /// Defines the 5th AGC Threshold.
  agc_step_5 { 0 4 }
}

spi_reg! {
  /// Control of the PLL bandwidth.
  Pll;
  NeverRegFskOok;
  FskOok;
  0x70 0xD0 RSpiReg WSpiReg;
  /// Controls the PLL bandwidth.
  pll_bandwidth { 6 2 }
}
