//! LoRa mode registers.

use drone_core::reg::RegVal;
use drone_sx1276_macros::spi_reg;
use mode::Lora;
use reg::{RSpiReg, SpiReg, WSpiReg};
use reg::never::NeverRegLora;

spi_reg! {
  /// Operating mode & LoRa™ / FSK selection.
  OpMode;
  NeverRegLora;
  Lora;
  0x01 0x81 RSpiReg WSpiReg;
  /// LoRa™ / FSK selection.
  long_range_mode { 7 1 }
  /// Allows access to FSK registers page located in address space (0x0D:0x3F)
  /// while in LoRa mode.
  access_shared_reg { 6 1 }
  /// Access Low Frequency Mode registers.
  low_frequency_mode_on { 3 1 }
  /// Transceiver mode.
  mode { 0 3 }
}

spi_reg! {
  /// RF Carrier Frequency, Most Significant Bits.
  FrMsb;
  FrMid;
  Lora;
  0x06 0x6C RSpiReg WSpiReg;
  /// MSB of RF carrier frequency.
  frf_23_16 { 0 8 }
}

spi_reg! {
  /// RF Carrier Frequency, Intermediate Bits.
  FrMid;
  FrLsb;
  Lora;
  0x07 0x80 RSpiReg WSpiReg;
  /// MSB of RF carrier frequency.
  frf_15_8 { 0 8 }
}

spi_reg! {
  /// RF Carrier Frequency, Least Significant Bits.
  FrLsb;
  PaConfig;
  Lora;
  0x08 0x00 RSpiReg WSpiReg;
  /// LSB of RF carrier frequency.
  frf_7_0 { 0 8 }
}

spi_reg! {
  /// PA selection and Output Power control.
  PaConfig;
  PaRamp;
  Lora;
  0x09 0x4F RSpiReg WSpiReg;
  /// Selects PA output pin.
  pa_select { 7 1 }
  /// Select max output power.
  max_power { 4 3 }
  /// Select output power.
  output_power { 0 4 }
}

spi_reg! {
  /// Control of PA ramp time, low phase noise PLL.
  PaRamp;
  Ocp;
  Lora;
  0x0A 0x09 RSpiReg WSpiReg;
  /// Rise/Fall time of ramp up/down in FSK.
  pa_ramp { 0 4 }
}

spi_reg! {
  /// Over Current Protection control.
  Ocp;
  Lna;
  Lora;
  0x0B 0x2B RSpiReg WSpiReg;
  /// Enables overload current protection (OCP) for the PA.
  ocp_on { 5 1 }
  /// Trimming of OCP current.
  ocp_trim { 0 5 }
}

spi_reg! {
  /// LNA settings.
  Lna;
  FifoAddrPtr;
  Lora;
  0x0C 0x20 RSpiReg WSpiReg;
  /// LNA gain setting.
  lna_gain { 5 3 }
  /// Low Frequency (RFI_LF) LNA current adjustment.
  lna_boost_lf { 3 2 }
  /// High Frequency (RFI_HF) LNA current adjustment.
  lna_boost_hf { 0 2 }
}

spi_reg! {
  /// FIFO SPI pointer.
  FifoAddrPtr;
  FifoTxBaseAddr;
  Lora;
  0x0D 0x00 RSpiReg WSpiReg;
  /// SPI interface address pointer in FIFO data buffer.
  fifo_addr_ptr { 0 8 }
}

spi_reg! {
  /// Start Tx data.
  FifoTxBaseAddr;
  FifoRxBaseAddr;
  Lora;
  0x0E 0x80 RSpiReg WSpiReg;
  /// Write base address in FIFO data buffer for TX modulator.
  fifo_tx_base_addr { 0 8 }
}

spi_reg! {
  /// Start Rx data.
  FifoRxBaseAddr;
  FifoRxCurrentAddr;
  Lora;
  0x0F 0x00 RSpiReg WSpiReg;
  /// Read base address in FIFO data buffer for RX demodulator.
  fifo_rx_base_addr { 0 8 }
}

spi_reg! {
  /// Start address of last packet received.
  FifoRxCurrentAddr;
  IrqFlagsMask;
  Lora;
  0x10 0 RSpiReg;
  /// Start address (in data buffer) of last packet received.
  fifo_rx_current_addr { 0 8 }
}

spi_reg! {
  /// Optional IRQ flag mask.
  IrqFlagsMask;
  IrqFlags;
  Lora;
  0x11 0x00 RSpiReg WSpiReg;
  /// Timeout interrupt mask.
  rx_timeout_mask { 7 1 }
  /// Packet reception complete interrupt mask.
  rx_done_mask { 6 1 }
  /// Payload CRC error interrupt mask.
  payload_crc_error_mask { 5 1 }
  /// Valid header received in Rx mask.
  valid_header_mask { 4 1 }
  /// FIFO Payload transmission complete interrupt mask.
  tx_done_mask { 3 1 }
  /// CAD complete interrupt mask.
  cad_done_mask { 2 1 }
  /// FHSS change channel interrupt mask.
  fhss_change_channel_mask { 1 1 }
  /// Cad Detected Interrupt Mask.
  cad_detected_mask { 0 1 }
}

spi_reg! {
  /// IRQ flags.
  IrqFlags;
  RxNbBytes;
  Lora;
  0x12 0x00 RSpiReg WSpiReg;
  /// Timeout interrupt.
  rx_timeout { 7 1 }
  /// Packet reception complete interrupt.
  rx_done { 6 1 }
  /// Payload CRC error interrupt.
  payload_crc_error { 5 1 }
  /// Valid header received in Rx.
  valid_header { 4 1 }
  /// FIFO Payload transmission complete interrupt.
  tx_done { 3 1 }
  /// CAD complete interrupt.
  cad_done { 2 1 }
  /// FHSS change channel interrupt.
  fhss_change_channel { 1 1 }
  /// Valid Lora signal detected during CAD operation.
  cad_detected { 0 1 }
}

spi_reg! {
  /// Number of received bytes.
  RxNbBytes;
  RxHeaderCntValueMsb;
  Lora;
  0x13 0 RSpiReg;
  /// Number of payload bytes of latest packet received.
  fifo_rx_bytes_nb { 0 8 }
}

spi_reg! {
  /// Number of valid headers received.
  RxHeaderCntValueMsb;
  RxHeaderCntValueLsb;
  Lora;
  0x14 0 RSpiReg;
  /// Number of valid headers received since last transition into Rx mode,
  /// MSB(15:8).
  valid_header_cnt_msb_15_8 { 0 8 }
}

spi_reg! {
  /// Number of valid headers received.
  RxHeaderCntValueLsb;
  RxPacketCntValueMsb;
  Lora;
  0x15 0 RSpiReg;
  /// Number of valid headers received since last transition into Rx mode,
  /// LSB(7:0).
  valid_header_cnt_lsb_7_0 { 0 8 }
}

spi_reg! {
  /// Number of valid packets received.
  RxPacketCntValueMsb;
  RxPacketCntValueLsb;
  Lora;
  0x16 0 RSpiReg;
  /// Number of valid packets received since last transition into Rx mode,
  /// MSB(15:8).
  valid_packet_cnt_msb_15_8 { 0 8 }
}

spi_reg! {
  /// Number of valid packets received.
  RxPacketCntValueLsb;
  ModemStat;
  Lora;
  0x17 0 RSpiReg;
  /// Number of valid packets received since last transition into Rx mode,
  /// LSB(7:0).
  valid_packet_cnt_lsb_7_0 { 0 8 }
}

spi_reg! {
  /// Live LoRa™ modem status.
  ModemStat;
  PktSnrValue;
  Lora;
  0x18 0 RSpiReg;
  /// Coding rate of last header received.
  rx_coding_rate { 5 3 }
  /// Modem clear.
  modem_clear { 4 1 }
  /// Header info valid.
  header_info_valid { 3 1 }
  /// RX on-going.
  rx_ongoing { 2 1 }
  /// Signal synchronized.
  signal_synchronized { 1 1 }
  /// Signal detected.
  signal_detected { 0 1 }
}

spi_reg! {
  /// Estimation of last packet SNR.
  PktSnrValue;
  PktRssiValue;
  Lora;
  0x19 0 RSpiReg;
  /// Estimation of SNR on last packet received. In two's compliment format
  /// multiplied by 4.
  packet_snr { 0 8 }
}

spi_reg! {
  /// RSSI of last packet.
  PktRssiValue;
  RssiValue;
  Lora;
  0x1A 0 RSpiReg;
  /// RSSI of the latest packet received (dBm).
  packet_rssi { 0 8 }
}

spi_reg! {
  /// Current RSSI.
  RssiValue;
  HopChannel;
  Lora;
  0x1B 0 RSpiReg;
  /// Current RSSI value (dBm).
  rssi { 0 8 }
}

spi_reg! {
  /// FHSS start channel.
  HopChannel;
  ModemConfig1;
  Lora;
  0x1C 0 RSpiReg;
  /// PLL failed to lock while attempting a TX/RX/CAD operation.
  pll_timeout { 7 1 }
  /// CRC Information extracted from the received packet header.
  crc_on_payload { 6 1 }
  /// Current value of frequency hopping channel in use.
  fhss_present_channel { 0 6 }
}

spi_reg! {
  /// Modem PHY config 1.
  ModemConfig1;
  ModemConfig2;
  Lora;
  0x1D 0x72 RSpiReg WSpiReg;
  /// Signal bandwidth.
  bw { 4 4 }
  /// Error coding rate.
  coding_rate { 1 3 }
  /// Enable Implicit Header mode.
  implicit_header_mode_on { 0 1 }
}

spi_reg! {
  /// Modem PHY config 2.
  ModemConfig2;
  SymbTimeoutLsb;
  Lora;
  0x1E 0x70 RSpiReg WSpiReg;
  /// SF rate (expressed as a base-2 logarithm).
  spreading_factor { 4 4 }
  /// Continuous mode select.
  tx_continuous_mode { 3 1 }
  /// Enable CRC generation and check on payload.
  rx_payload_crc_on { 2 1 }
  /// RX Time-Out MSB.
  symb_timeout_9_8 { 0 2 }
}

spi_reg! {
  /// Receiver timeout value.
  SymbTimeoutLsb;
  PreambleMsb;
  Lora;
  0x1F 0x64 RSpiReg WSpiReg;
  /// RX Time-Out LSB.
  symb_timeout_7_0 { 0 8 }
}

spi_reg! {
  /// Size of preamble.
  PreambleMsb;
  PreambleLsb;
  Lora;
  0x20 0x00 RSpiReg WSpiReg;
  /// Preamble length MSB.
  preamble_length_15_8 { 0 8 }
}

spi_reg! {
  /// Size of preamble.
  PreambleLsb;
  PayloadLength;
  Lora;
  0x21 0x08 RSpiReg WSpiReg;
  /// Preamble Length LSB.
  preamble_length_7_0 { 0 8 }
}

spi_reg! {
  /// LoRa™ payload length.
  PayloadLength;
  MaxPayloadLength;
  Lora;
  0x22 0x1 RSpiReg WSpiReg;
  /// Payload length in bytes.
  payload_length { 0 8 }
}

spi_reg! {
  /// LoRa™ maximum payload length.
  MaxPayloadLength;
  HopPeriod;
  Lora;
  0x23 0xFF RSpiReg WSpiReg;
  /// Maximum payload length.
  payload_max_length { 0 8 }
}

spi_reg! {
  /// FHSS Hop period.
  HopPeriod;
  FifoRxByteAddr;
  Lora;
  0x24 0x00 RSpiReg WSpiReg;
  /// Symbol periods between frequency hops.
  freq_hopping_period { 0 8 }
}

spi_reg! {
  /// Address of last byte written in FIFO.
  FifoRxByteAddr;
  ModemConfig3;
  Lora;
  0x25 0 RSpiReg;
  /// Current value of RX databuffer pointer (address of last byte written by
  /// Lora receiver).
  fifo_rx_byte_addr_ptr { 0 8 }
}

spi_reg! {
  /// Modem PHY config 3.
  ModemConfig3;
  PpmCorrection;
  Lora;
  0x26 0x00 RSpiReg WSpiReg;
  /// Enable Low Data Rate Optimize.
  low_data_rate_optimize { 3 1 }
  /// AGC.
  agc_auto_on { 2 1 }
}

spi_reg! {
  /// Data rate offset value.
  PpmCorrection;
  FeiMsb;
  Lora;
  0x27 0x00 RSpiReg WSpiReg;
  /// Data rate offset value, used in conjunction with AFC.
  ppm_correction { 0 8 }
}

spi_reg! {
  /// Estimated frequency error.
  FeiMsb;
  FeiMid;
  Lora;
  0x28 0x00 RSpiReg;
  /// MSB of RF Frequency Error.
  freq_error_19_16 { 0 4 }
}

spi_reg! {
  /// Estimated frequency error.
  FeiMid;
  FeiLsb;
  Lora;
  0x29 0x00 RSpiReg;
  /// Middle byte of RF Frequency Error.
  freq_error_15_8 { 0 8 }
}

spi_reg! {
  /// Estimated frequency error.
  FeiLsb;
  NeverRegLora;
  Lora;
  0x2A 0x00 RSpiReg;
  /// LSB of RF Frequency Error.
  freq_error_7_0 { 0 8 }
}

spi_reg! {
  /// Wideband RSSI measurement.
  RssiWideband;
  NeverRegLora;
  Lora;
  0x2C 0 RSpiReg;
  /// Wideband RSSI measurement used to locally generate a random number.
  rssi_wideband_7_0 { 0 8 }
}

spi_reg! {
  /// LoRa detection Optimize for SF6.
  DetectOptimize;
  NeverRegLora;
  Lora;
  0x31 0xC3 RSpiReg WSpiReg;
  /// LoRa Detection Optimize.
  detection_optimize { 0 3 }
}

spi_reg! {
  /// Invert LoRa I and Q signals.
  InvertIQ;
  NeverRegLora;
  Lora;
  0x33 0x27 RSpiReg WSpiReg;
  /// Invert the LoRa I and Q signals.
  invert_i_q { 6 1 }
}

spi_reg! {
  /// LoRa detection threshold for SF6.
  DetectionThreshold;
  NeverRegLora;
  Lora;
  0x37 0x0A RSpiReg WSpiReg;
  /// LoRa detection threshold.
  detection_threshold { 0 8 }
}

spi_reg! {
  /// LoRa Sync Word.
  SyncWord;
  NeverRegLora;
  Lora;
  0x39 0x12 RSpiReg WSpiReg;
  /// LoRa Sync Word.
  sync_word { 0 8 }
}
