//! Semtech SX1276/77/78/79 communication session.
//!
//! See [`Sess`] documentation for more details.
//!
//! [`Sess`]: struct.Sess.html

pub mod recv;
pub mod send;

use conn::{self, Conn};
use core::fmt;
use device::Device;
use mode::{FskOok, Lora, Mode, Undefined};

/// Semtech SX1276/77/78/79 communication session.
pub struct Sess<C: Conn, M: Mode> {
  /// A Semtech SX1276/77/78/79 device,
  pub device: Device<C, M>,
  /// A SPI connection interface.
  pub spi: C::Spi,
  /// A byte buffer for I/O operations.
  pub buffer: Vec<u8>,
}

/// The error returned from [`Sess::store`].
///
/// [`Sess::store`]: struct.Sess.html#method.store
pub struct SendError<C: Conn> {
  /// A communication session.
  pub sess: Sess<C, Undefined>,
  /// A kind of the error.
  pub kind: C::SendErrorKind,
}

impl<C: Conn, M: Mode> Sess<C, M> {
  /// Creates a new [`Sess`].
  ///
  /// See [`Sess`] for the arguments description.
  ///
  /// [`Sess`]: struct.Sess.html
  #[inline(always)]
  pub fn new(device: Device<C, M>, spi: C::Spi, buffer: Vec<u8>) -> Self {
    Self {
      device,
      spi,
      buffer,
    }
  }

  /// Destructures the `Sess`.
  #[inline(always)]
  pub fn destroy(self) -> (Device<C, M>, C::Spi, Vec<u8>) {
    let Self {
      device,
      spi,
      buffer,
    } = self;
    (device, spi, buffer)
  }

  /// Writes `data` bytes at the `address`.
  pub unsafe fn send_raw(
    self,
    address: u8,
    data: &[u8],
  ) -> impl Future<Item = Self, Error = SendError<C>> {
    let (device, spi, mut buffer) = self.destroy();
    assert!(buffer.len() > data.len());
    let future = device.into_tokens().transfer(
      spi,
      address | 0x80,
      data.len(),
      data.as_ptr(),
      buffer.as_mut_ptr(),
    );
    compose_session(future, buffer)
  }

  /// Reads `count` bytes from the `address`.
  pub unsafe fn recv_raw(
    self,
    address: u8,
    count: usize,
  ) -> impl Future<Item = Self, Error = SendError<C>> {
    let (device, spi, mut buffer) = self.destroy();
    assert!(buffer.len() > count);
    let future = device.into_tokens().transfer(
      spi,
      address,
      count,
      buffer[1..=count].as_ptr(),
      buffer.as_mut_ptr(),
    );
    compose_session(future, buffer)
  }

  /// Writes bytes to the FIFO.
  #[inline(always)]
  pub fn send_data(
    self,
    data: &[u8],
  ) -> impl Future<Item = Self, Error = SendError<C>> {
    unsafe { self.send_raw(0x00, data) }
  }

  /// Reads bytes from the FIFO.
  #[inline(always)]
  pub fn recv_data(
    self,
    count: usize,
  ) -> impl Future<Item = Self, Error = SendError<C>> {
    unsafe { self.recv_raw(0x00, count) }
  }
}

impl<C: Conn> Sess<C, FskOok> {
  /// Converts into LoRa mode.
  #[inline(always)]
  pub unsafe fn into_lora(self) -> Sess<C, Lora> {
    let (device, spi, buffer) = self.destroy();
    Sess::new(device.into_lora(), spi, buffer)
  }
}

impl<C: Conn> Sess<C, Lora> {
  /// Converts into FSK/OOK mode.
  #[inline(always)]
  pub unsafe fn into_fsk_ook(self) -> Sess<C, FskOok> {
    let (device, spi, buffer) = self.destroy();
    Sess::new(device.into_fsk_ook(), spi, buffer)
  }
}

impl<C: Conn> Sess<C, Undefined> {
  /// Converts into LoRa mode.
  #[inline(always)]
  pub unsafe fn into_lora(self) -> Sess<C, Lora> {
    let (device, spi, buffer) = self.destroy();
    Sess::new(device.into_lora(), spi, buffer)
  }

  /// Converts into FSK/OOK mode.
  #[inline(always)]
  pub unsafe fn into_fsk_ook(self) -> Sess<C, FskOok> {
    let (device, spi, buffer) = self.destroy();
    Sess::new(device.into_fsk_ook(), spi, buffer)
  }
}

impl<C: Conn> fmt::Debug for SendError<C> {
  #[inline(always)]
  fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
    self.kind.fmt(f)
  }
}

impl<C: Conn> PartialEq for SendError<C> {
  #[inline(always)]
  fn eq(&self, other: &Self) -> bool {
    self.kind.eq(&other.kind)
  }
}

impl<C: Conn> Eq for SendError<C> {}

#[cfg_attr(feature = "clippy", allow(boxed_local))]
#[inline(always)]
fn compose_session<C, M, F>(
  future: Box<F>,
  buffer: Vec<u8>,
) -> impl Future<Item = Sess<C, M>, Error = SendError<C>>
where
  C: Conn,
  M: Mode,
  F: Future<Item = (C, C::Spi), Error = conn::SendError<C>>,
  F: ?Sized,
{
  AsyncFuture::new(move || match await!(future) {
    Ok((tokens, spi)) => {
      Ok(Sess::new(Device::from_tokens(tokens), spi, buffer))
    }
    Err(conn::SendError { tokens, spi, kind }) => Err(SendError {
      sess: Sess::new(Device::from_tokens(tokens), spi, buffer),
      kind,
    }),
  })
}
