//! Typed SPI read.

use conn::Conn;
use mode::Mode;
use reg::RSpiReg;
use sess::{SendError, Sess};

macro_rules! sess_recv {
  (
    $doc:expr,
    $name:ident,
    ($($r:ident[$ri:expr]: $r_ty:ident -> ($($r_next:ident),*)),*),
  ) => {
    #[doc = $doc]
    pub fn $name<$($r_ty),*>(
      self,
    ) -> impl Future<Item = (Self, $($r_ty),*), Error = SendError<C>>
    where
      $($r_ty: RSpiReg<M $(, Next = $r_next)*>,)*
    {
      let sess = unsafe { self.recv_raw(R1::ADDRESS, 1) };
      AsyncFuture::new(move || {
        let sess = await!(sess)?;
        $(
          let $r = unsafe { $r_ty::from_raw(*sess.buffer.get_unchecked($ri)) };
        )*
        Ok((sess $(, $r)*))
      })
    }
  }
}

impl<C: Conn, M: Mode> Sess<C, M> {
  sess_recv! {
    "Reads a register.",
    recv,
    (r1[1]: R1 -> ()),
  }

  sess_recv! {
    "Reads 2 registers in burst mode.",
    recv2,
    (r1[1]: R1 -> (R2), r2[2]: R2 -> ()),
  }

  sess_recv! {
    "Reads 3 registers in burst mode.",
    recv3,
    (r1[1]: R1 -> (R2), r2[2]: R2 -> (R3), r3[3]: R3 -> ()),
  }

  sess_recv! {
    "Reads 4 registers in burst mode.",
    recv4,
    (r1[1]: R1 -> (R2), r2[2]: R2 -> (R3), r3[3]: R3 -> (R4), r4[4]: R4 -> ()),
  }

  sess_recv! {
    "Reads 5 registers in burst mode.",
    recv5,
    (r1[1]: R1 -> (R2), r2[2]: R2 -> (R3), r3[3]: R3 -> (R4), r4[4]: R4 -> (R5),
     r5[5]: R5 -> ()),
  }

  sess_recv! {
    "Reads 6 registers in burst mode.",
    recv6,
    (r1[1]: R1 -> (R2), r2[2]: R2 -> (R3), r3[3]: R3 -> (R4), r4[4]: R4 -> (R5),
     r5[5]: R5 -> (R6), r6[6]: R6 -> ()),
  }

  sess_recv! {
    "Reads 7 registers in burst mode.",
    recv7,
    (r1[1]: R1 -> (R2), r2[2]: R2 -> (R3), r3[3]: R3 -> (R4), r4[4]: R4 -> (R5),
     r5[5]: R5 -> (R6), r6[6]: R6 -> (R7), r7[7]: R7 -> ()),
  }

  sess_recv! {
    "Reads 8 registers in burst mode.",
    recv8,
    (r1[1]: R1 -> (R2), r2[2]: R2 -> (R3), r3[3]: R3 -> (R4), r4[4]: R4 -> (R5),
     r5[5]: R5 -> (R6), r6[6]: R6 -> (R7), r7[7]: R7 -> (R8), r8[8]: R8 -> ()),
  }
}
