//! Typed SPI write.

use conn::Conn;
use drone_core::reg::RegVal;
use mode::{FskOok, Lora, Mode};
use reg::{fsk_ook, lora, SpiReg, WSpiReg};
use sess::{SendError, Sess};

macro_rules! sess_send {
  (
    $doc:expr,
    $name:ident,
    $op_mode_doc:expr,
    $op_mode_name:ident,
    $fn_doc:expr,
    $fn_name:ident,
    $op_mode_fn_doc:expr,
    $op_mode_fn_name:ident,
    ($($head:ident)*) ->
    ($($r:ident -> $r_next:ident),*) ->
    ($($tail:ident)*),
    ($($f:ident: $f_ty:ident -> $r_ty:ident),*),
  ) => {
    #[doc = $doc]
    pub trait $name<F1 $(,$f_ty)*, C: Conn, M: Mode>: Sized {
      #[doc = $fn_doc]
      #[cfg_attr(feature = "clippy", allow(too_many_arguments))]
      fn $fn_name<R1 $(,$r_ty)*>(
        self,
        f1: F1,
        $($f: $f_ty,)*
      ) -> Box<Future<Item = Self, Error = SendError<C>> + Send>
      where
        R1: WSpiReg<M $(, Next = $head)*>,
        $($r: WSpiReg<M, Next = $r_next>,)*
        $($tail: WSpiReg<M>,)*
        F1: FnOnce(&mut R1) -> &mut R1,
        $($f_ty: FnOnce(&mut $r_ty) -> &mut $r_ty,)*;
    }

    #[doc = $op_mode_doc]
    pub trait $op_mode_name<F1 $(,$f_ty)*, C: Conn, M: Mode>: Sized {
      /// Resulting session type.
      type Output;

      /// Concrete `OpMode` register.
      type OpMode: WSpiReg<M>;

      #[doc = $op_mode_fn_doc]
      #[cfg_attr(feature = "clippy", allow(too_many_arguments))]
      fn $op_mode_fn_name<$($r_ty),*>(
        self,
        f1: F1,
        $($f: $f_ty,)*
      ) -> Box<Future<Item = Self::Output, Error = SendError<C>> + Send>
      where
        $($r: WSpiReg<M, Next = $r_next>,)*
        $($tail: WSpiReg<M>,)*
        F1: FnOnce(&mut Self::OpMode) -> &mut Self::OpMode,
        $($f_ty: FnOnce(&mut $r_ty) -> &mut $r_ty,)*;
    }

    impl<F1 $(,$f_ty)*, C, M> $name<F1 $(,$f_ty)*, C, M> for Sess<C, M>
    where
      C: Conn,
      M: Mode,
    {
      #[inline(always)]
      fn $fn_name<R1 $(,$r_ty)*>(
        self,
        f1: F1,
        $($f: $f_ty,)*
      ) -> Box<Future<Item = Self, Error = SendError<C>> + Send>
      where
        R1: WSpiReg<M $(, Next = $head)*>,
        $($r: WSpiReg<M, Next = $r_next>,)*
        $($tail: WSpiReg<M>,)*
        F1: FnOnce(&mut R1) -> &mut R1,
        $($f_ty: FnOnce(&mut $r_ty) -> &mut $r_ty,)*
      {
        unsafe {
          Box::new(self.send_raw(
            R1::ADDRESS,
            &[
              f1(&mut R1::default()).raw(),
              $($f(&mut $r_ty::default()).raw()),*
            ],
          ))
        }
      }
    }

    impl<F1 $(,$f_ty)*, C> $op_mode_name<F1 $(,$f_ty)*, C, Lora>
      for Sess<C, FskOok>
    where
      C: Conn,
    {
      type Output = Sess<C, Lora>;
      type OpMode = lora::OpMode;

      #[inline(always)]
      fn $op_mode_fn_name<$($r_ty),*>(
        self,
        f1: F1,
        $($f: $f_ty,)*
      ) -> Box<Future<Item = Self::Output, Error = SendError<C>> + Send>
      where
        $($r: WSpiReg<Lora, Next = $r_next>,)*
        $($tail: WSpiReg<Lora>,)*
        F1: FnOnce(&mut Self::OpMode) -> &mut Self::OpMode,
        $($f_ty: FnOnce(&mut $r_ty) -> &mut $r_ty,)*
      {
        unsafe {
          Box::new(self.send_raw(
            lora::OpMode::ADDRESS,
            &[
              f1(&mut lora::OpMode::default()).raw(),
              $($f(&mut $r_ty::default()).raw()),*
            ],
          ).map(|sess| sess.into_lora()))
        }
      }
    }

    impl<F1 $(,$f_ty)*, C> $op_mode_name<F1 $(,$f_ty)*, C, FskOok>
      for Sess<C, Lora>
    where
      C: Conn,
    {
      type Output = Sess<C, FskOok>;
      type OpMode = fsk_ook::OpMode;

      #[inline(always)]
      fn $op_mode_fn_name<$($r_ty),*>(
        self,
        f1: F1,
        $($f: $f_ty,)*
      ) -> Box<Future<Item = Self::Output, Error = SendError<C>> + Send>
      where
        $($r: WSpiReg<FskOok, Next = $r_next>,)*
        $($tail: WSpiReg<FskOok>,)*
        F1: FnOnce(&mut Self::OpMode) -> &mut Self::OpMode,
        $($f_ty: FnOnce(&mut $r_ty) -> &mut $r_ty,)*
      {
        unsafe {
          Box::new(self.send_raw(
            fsk_ook::OpMode::ADDRESS,
            &[
              f1(&mut fsk_ook::OpMode::default()).raw(),
              $($f(&mut $r_ty::default()).raw()),*
            ],
          ).map(|sess| sess.into_fsk_ook()))
        }
      }
    }
  }
}

sess_send! {
  "A session trait for writing a register.",
  SessSend1,
  "A session trait for writing to the op mode register.",
  SessSend1OpMode,
  "Writes a register.",
  send,
  "Writes to the op mode register.",
  send_op_mode,
  () -> () -> (),
  (),
}

sess_send! {
  "A session trait for writing 2 registers in burst mode.",
  SessSend2,
  "A session trait for writing to the op mode and another register in burst
  mode.",
  SessSend2OpMode,
  "Writes 2 registers in burst mode.",
  send2,
  "Writes to the op mode and another register in burst mode.",
  send2_op_mode,
  (R2) -> () -> (R2),
  (f2: F2 -> R2),
}

sess_send! {
  "A session trait for writing 3 registers in burst mode.",
  SessSend3,
  "A session trait for writing to the op mode and 2 other registers in burst
  mode.",
  SessSend3OpMode,
  "Writes 3 registers in burst mode.",
  send3,
  "Writes to the op mode and 2 other registers in burst mode.",
  send3_op_mode,
  (R2) -> (R2 -> R3) -> (R3),
  (f2: F2 -> R2, f3: F3 -> R3),
}

sess_send! {
  "A session trait for writing 4 registers in burst mode.",
  SessSend4,
  "A session trait for writing to the op mode and 3 other registers in burst
  mode.",
  SessSend4OpMode,
  "Writes 4 registers in burst mode.",
  send4,
  "Writes to the op mode and 3 other registers in burst mode.",
  send4_op_mode,
  (R2) -> (R2 -> R3, R3 -> R4) -> (R4),
  (f2: F2 -> R2, f3: F3 -> R3, f4: F4 -> R4),
}

sess_send! {
  "A session trait for writing 5 registers in burst mode.",
  SessSend5,
  "A session trait for writing to the op mode and 4 other registers in burst
  mode.",
  SessSend5OpMode,
  "Writes 5 registers in burst mode.",
  send5,
  "Writes to the op mode and 4 other registers in burst mode.",
  send5_op_mode,
  (R2) -> (R2 -> R3, R3 -> R4, R4 -> R5) -> (R5),
  (f2: F2 -> R2, f3: F3 -> R3, f4: F4 -> R4, f5: F5 -> R5),
}

sess_send! {
  "A session trait for writing 6 registers in burst mode.",
  SessSend6,
  "A session trait for writing to the op mode and 5 other registers in burst
  mode.",
  SessSend6OpMode,
  "Writes 6 registers in burst mode.",
  send6,
  "Writes to the op mode and 5 other registers in burst mode.",
  send6_op_mode,
  (R2) -> (R2 -> R3, R3 -> R4, R4 -> R5, R5 -> R6) -> (R6),
  (f2: F2 -> R2, f3: F3 -> R3, f4: F4 -> R4, f5: F5 -> R5, f6: F6 -> R6),
}

sess_send! {
  "A session trait for writing 7 registers in burst mode.",
  SessSend7,
  "A session trait for writing to the op mode and 6 other registers in burst
  mode.",
  SessSend7OpMode,
  "Writes 7 registers in burst mode.",
  send7,
  "Writes to the op mode and 6 other registers in burst mode.",
  send7_op_mode,
  (R2) -> (R2 -> R3, R3 -> R4, R4 -> R5, R5 -> R6, R6 -> R7) -> (R7),
  (f2: F2 -> R2, f3: F3 -> R3, f4: F4 -> R4, f5: F5 -> R5, f6: F6 -> R6,
   f7: F7 -> R7),
}

sess_send! {
  "A session trait for writing 8 registers in burst mode.",
  SessSend8,
  "A session trait for writing to the op mode and 7 other registers in burst
  mode.",
  SessSend8OpMode,
  "Writes 8 registers in burst mode.",
  send8,
  "Writes to the op mode and 7 other registers in burst mode.",
  send8_op_mode,
  (R2) -> (R2 -> R3, R3 -> R4, R4 -> R5, R5 -> R6, R6 -> R7, R7 -> R8) -> (R8),
  (f2: F2 -> R2, f3: F3 -> R3, f4: F4 -> R4, f5: F5 -> R5, f6: F6 -> R6,
   f7: F7 -> R7, f8: F8 -> R8),
}
